﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Navigation;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;
using Demo03.WindowsPhone.Resources;
using System.Windows.Media;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;
using Windows.Networking.Proximity;
using System.Threading.Tasks;

namespace Demo03.WindowsPhone
{
    public partial class MainPage : PhoneApplicationPage
    {
        private bool isConnected;

        private StreamSocket socket;
        private DataReader reader;
        private DataWriter writer;

        // Constructor
        public MainPage()
        {
            InitializeComponent();
        }

        protected override void OnNavigatedTo(NavigationEventArgs e)
        {
            // Controlla se il device supporta la modalità tap-to-connect.
            if (PeerFinder.SupportedDiscoveryTypes.HasFlag(PeerDiscoveryTypes.Triggered))
            {
                PeerFinder.TriggeredConnectionStateChanged += PeerFinder_TriggeredConnectionStateChanged;
                PeerFinder.Start();
            }
            else
            {
                statusTextBlock.Text = "Modalità tap-to-connect non supportata.";
            }

            base.OnNavigatedTo(e);
        }

        protected override void OnNavigatedFrom(NavigationEventArgs e)
        {
            // Quando si esce dalla pagina, chiude gli eventuali socket aperti.
            this.CloseSocket();

            PeerFinder.Stop();
            PeerFinder.TriggeredConnectionStateChanged -= PeerFinder_TriggeredConnectionStateChanged;

            base.OnNavigatedFrom(e);
        }

        private void PeerFinder_TriggeredConnectionStateChanged(object sender, TriggeredConnectionStateChangedEventArgs args)
        {
            Dispatcher.BeginInvoke(() =>
            {
                statusTextBlock.Text = args.State.ToString();

                switch (args.State)
                {
                    case TriggeredConnectState.PeerFound:   // Il peer è stato trovato, è possibile allontanare i device.
                        statusTextBlock.Text += ". Peer identificato";
                        break;

                    case TriggeredConnectState.Listening:   // In ascolto.
                        break;

                    case TriggeredConnectState.Connecting:  // Creazione della connessione via Socket.
                        break;

                    case TriggeredConnectState.Completed:   // La connessione è stabilita ed è disponibile.
                        isConnected = true;
                        sendButton.IsEnabled = true;

                        PeerFinder.Stop();
                        statusTextBlock.Text += ". Connessione stabilita";

                        this.HandleConnection(args.Socket);
                        break;

                    case TriggeredConnectState.Canceled:    // La connessione è stata annulla.
                        break;

                    case TriggeredConnectState.Failed:  // Impossibile stabilire la connessione.
                        break;

                    default:
                        break;
                }
            });
        }

        private void HandleConnection(StreamSocket streamSocket)
        {
            // Configura il socket.
            socket = streamSocket;
            reader = new DataReader(streamSocket.InputStream) { InputStreamOptions = InputStreamOptions.None };
            writer = new DataWriter(streamSocket.OutputStream);

            this.ListenAsync(reader);
        }

        private async void ListenAsync(DataReader socketReader)
        {
            try
            {
                // Controlla se lo stream contiene dati, verificando la presenza di un numero che contiene
                // la lunghezza del testo.
                var bytesRead = await socketReader.LoadAsync(sizeof(uint));
                if (bytesRead > 0)
                {
                    // Legge il numero di caratteri presenti nello stream e, subito dopo, carica dallo stream
                    // il numero di byte specificati.
                    var len = socketReader.ReadUInt32();
                    bytesRead = await socketReader.LoadAsync(len);
                    if (bytesRead > 0)
                    {
                        // Se effettivamente ci sono dati, li legge in formato stringa.
                        var message = socketReader.ReadString(len);

                        // Richiama il metodo che gestisce il messaggio ricevuto.
                        var continueRead = this.HandleMessage(message);

                        if (continueRead)
                        {
                            // Sono previsti ulteriori messaggi, quindi avvia una nuova lettura.
                            this.ListenAsync(socketReader);
                        }
                        else
                        {
                            // La connessione è terminata.
                            this.CloseSocket();
                        }
                    }
                    else
                    {
                        // Il socket remoto è stato chiuso.
                        this.CloseSocket();
                    }
                }
                else
                {
                    this.CloseSocket();
                }
            }
            catch
            {
                this.CloseSocket();
            }
        }

        private bool HandleMessage(string message)
        {
            this.InsertPartnerMessage(message);
            return true;
        }

        private async Task SendAsync(string message)
        {
            if (isConnected && !string.IsNullOrEmpty(message))
            {
                try
                {
                    // Calcola il numero di byte da inviare.
                    var len = writer.MeasureString(message);
                    writer.WriteUInt32(len);

                    // Invia il messaggio vero e proprio.
                    writer.WriteString(message);
                    var bytesWritten = await writer.StoreAsync();
                    if (bytesWritten == 0)
                    {
                        // Il socket remoto è stato disconnesso.
                        this.CloseSocket();
                    }
                }
                catch
                {
                    this.CloseSocket();
                }
            }
            else
            {
                this.CloseSocket();
            }
        }

        private void CloseSocket()
        {
            if (reader != null)
            {
                reader.Dispose();
                reader = null;
            }

            if (writer != null)
            {
                writer.Dispose();
                writer = null;
            }

            if (socket != null)
            {
                socket.Dispose();
                socket = null;
            }

            isConnected = false;

            // Chiuso il socket, si mette in attesa di una connessione.
            statusTextBlock.Text = "Avvicina il peer per effettuare la connessione";
            PeerFinder.Start();
        }

        private async void sendButton_Click(object sender, RoutedEventArgs e)
        {
            if (!string.IsNullOrWhiteSpace(messageTextBox.Text))
            {
                await this.SendAsync(messageTextBox.Text);
                this.InsertMyMessage(messageTextBox.Text);
                messageTextBox.Text = string.Empty;
            }
        }

        private void InsertPartnerMessage(string text)
        {
            var item = new ListBoxItem
            {
                HorizontalAlignment = HorizontalAlignment.Left,
                Margin = new Thickness(20, 0, 0, 0),
                Foreground = new SolidColorBrush(Colors.Black)
            };

            item.Content = text;
            messagesListBox.Items.Add(item);
            this.ScrollToBottom();
        }

        private void InsertMyMessage(string text)
        {
            var item = new ListBoxItem
            {
                HorizontalAlignment = HorizontalAlignment.Right,
                Margin = new Thickness(0, 5, 20, 5),
                Foreground = new SolidColorBrush(Colors.Black)
            };

            item.Content = text;
            messagesListBox.Items.Add(item);
            this.ScrollToBottom();
        }

        private void ScrollToBottom()
        {
            var selectedIndex = messagesListBox.Items.Count - 1;
            if (selectedIndex < 0)
                return;

            messagesListBox.SelectedIndex = selectedIndex;
            messagesListBox.UpdateLayout();

            messagesListBox.ScrollIntoView(messagesListBox.SelectedItem);
            messagesListBox.SelectedIndex = -1;
        }

    }
}