﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Navigation;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;
using Windows.Networking.Proximity;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;
using System.Threading.Tasks;
using System.Diagnostics;

namespace EasyCheckin.WindowsPhone
{
    public partial class EventPage : PhoneApplicationPage
    {
        private bool isConnected;

        private StreamSocket socket;
        private DataReader reader;
        private DataWriter writer;

        private long eventID;
        private long registrationID;

        public EventPage()
        {
            InitializeComponent();
        }

        protected async override void OnNavigatedTo(NavigationEventArgs args)
        {
            eventID = long.Parse(NavigationContext.QueryString["eventID"]);

            // Recupera i dettagli dell'evento.
            var currentEvent = (await (from e in App.MobileService.GetTable<Event>()
                                       where e.Id == eventID
                                       select e).ToListAsync()).FirstOrDefault();

            // Mostra i dettagli dell'evento.
            eventDetails.DataContext = currentEvent;

            // Controlla se l'utente corrente è registrato all'evento.
            var registration = (await App.MobileService.GetTable<Registration>()
                .Where(r => r.EventId == eventID &&
                    r.UserId == App.MobileService.CurrentUser.UserId)
                .ToListAsync()).FirstOrDefault();

            if (registration == null)
            {
                // L'utente non è ancora registrato, quindi mostra il pulsante per effettuare la registrazione.
                registerButton.Visibility = System.Windows.Visibility.Visible;
            }
            else if (registration != null && !registration.Confirmed)
            {
                // L'utente è registrato, ma la registrazione non è confermata, quindi mostra il pulsante
                // per effettuare il check-in.
                registrationID = registration.Id;
                checkInButton.Visibility = System.Windows.Visibility.Visible;
            }
            else
            {
                // L'utente è già registrato e la registrazione è confermata, non sono possibili altre azioni.
                statusTextBlock.Text = "Check-in già effettuato";
            }

            base.OnNavigatedTo(args);
        }

        protected override void OnNavigatedFrom(NavigationEventArgs e)
        {
            // Quando si esce dalla pagina, chiude gli eventuali socket aperti.
            this.CloseSocket();

            PeerFinder.Stop();
            PeerFinder.TriggeredConnectionStateChanged -= PeerFinder_TriggeredConnectionStateChanged;
            statusTextBlock.Text = "Avvicina il telefono al tablet per confermare la registrazione";

            base.OnNavigatedFrom(e);
        }

        private async void registerButton_Click(object sender, RoutedEventArgs e)
        {
            // Effettua la registrazione all'evento.
            var registration = new Registration
            {
                EventId = eventID
                // Gli altri parametri sono inseriti dallo script lato server.
            };

            statusTextBlock.Text = "Registrazione in corso...";
            registerButton.IsEnabled = false;
            await App.MobileService.GetTable<Registration>().InsertAsync(registration);

            statusTextBlock.Text = "Registrazione effettuata con successo";
            registrationID = registration.Id;
            registerButton.Visibility = System.Windows.Visibility.Collapsed;
            checkInButton.Visibility = System.Windows.Visibility.Visible;
        }

        private void checkinButton_Click(object sender, RoutedEventArgs e)
        {
            // Controlla se il device supporta la modalità tap-to-connect.            
            if ((PeerFinder.SupportedDiscoveryTypes & PeerDiscoveryTypes.Triggered) != 0)
            {
                PeerFinder.TriggeredConnectionStateChanged += PeerFinder_TriggeredConnectionStateChanged;
                PeerFinder.Start();

                checkInButton.IsEnabled = false;
                statusTextBlock.Text = "Avvicina il telefono al tablet per confermare la registrazione";
            }
            else
            {
                MessageBox.Show("Modalità tap-to-connect non supportata.");
            }
        }

        private void PeerFinder_TriggeredConnectionStateChanged(object sender, 
            TriggeredConnectionStateChangedEventArgs e)
        {
            Dispatcher.BeginInvoke(async () =>
            {
                switch (e.State)
                {
                    case TriggeredConnectState.PeerFound:   // Il peer è stato trovato, è possibile allontanare i device.
                        statusTextBlock.Text = "Tablet identificato";
                        break;   
                    
                    case TriggeredConnectState.Listening:   // In ascolto.
                        break;

                    case TriggeredConnectState.Connecting:  // Creazione della connessione via Socket.
                        break;

                    case TriggeredConnectState.Completed:   // La connessione è stabilita ed è disponibile.
                        isConnected = true;
                        PeerFinder.Stop();
                        statusTextBlock.Text = "Connessione stabilita";
                        this.HandleConnection(e.Socket);

                        // Invia l'ID della registrazione al tablet.
                        await this.SendAsync(registrationID.ToString());

                        break;

                    case TriggeredConnectState.Canceled:    // La connessione è stata annulla.
                        break;

                    case TriggeredConnectState.Failed:  // Impossibile stabilire la connessione.
                        break;

                    default:
                        break;
                }
            });
        }

        private void HandleConnection(StreamSocket streamSocket)
        {
            // Configura il socket.
            socket = streamSocket;
            reader = new DataReader(streamSocket.InputStream) { InputStreamOptions = InputStreamOptions.None };
            writer = new DataWriter(streamSocket.OutputStream);

            this.ListenAsync(reader);
        }

        private async void ListenAsync(DataReader socketReader)
        {
            try
            {
                // Controlla se lo stream contiene dati, verificando la presenza di un numero che contiene
                // la lunghezza del testo.
                var bytesRead = await socketReader.LoadAsync(sizeof(uint));
                if (bytesRead > 0)
                {
                    // Legge il numero di caratteri presenti nello stream e, subito dopo, carica dallo stream
                    // il numero di byte specificati.
                    var len = socketReader.ReadUInt32();
                    bytesRead = await socketReader.LoadAsync(len);
                    if (bytesRead > 0)
                    {
                        // Se effettivamente ci sono dati, li legge in formato stringa.
                        var message = socketReader.ReadString(len);

                        // Richiama il metodo che gestisce il messaggio ricevuto.
                        var continueRead = this.HandleMessage(message);

                        if (continueRead)
                        {
                            // Sono previsti ulteriori messaggi, quindi avvia una nuova lettura.
                            this.ListenAsync(socketReader);
                        }
                        else
                        {
                            // La connessione è terminata.
                            this.CloseSocket();
                            statusTextBlock.Text = "Connessione terminata.";
                        }
                    }
                    else
                    {
                        // Il socket remoto è stato chiuso.
                        this.CloseSocket();
                    }
                }
                else
                {
                    this.CloseSocket();
                }
            }
            catch
            {
                this.CloseSocket();
            }
        }

        private bool HandleMessage(string message)
        {
            Dispatcher.BeginInvoke(() =>
            {
                statusTextBlock.Text = message;
            });

            return false;
        }

        private async Task SendAsync(string message)
        {
            if (isConnected && !string.IsNullOrEmpty(message))
            {
                try
                {
                    // Calcola il numero di byte da inviare.
                    var len = writer.MeasureString(message);
                    writer.WriteUInt32(len);

                    // Invia il messaggio vero e proprio.
                    writer.WriteString(message);
                    var bytesWritten = await writer.StoreAsync();
                    if (bytesWritten == 0)
                    {
                        // Il socket remoto è stato disconnesso.
                        this.CloseSocket();
                    }
                }
                catch
                {
                    this.CloseSocket();
                }
            }
            else
            {
                this.CloseSocket();
            }
        }

        private void CloseSocket()
        {
            if (reader != null)
            {
                reader.Dispose();
                reader = null;
            }

            if (writer != null)
            {
                writer.Dispose();
                writer = null;
            }

            if (socket != null)
            {
                socket.Dispose();
                socket = null;
            }

            isConnected = false;

            // Chiuso il socket, si mette in attesa di una connessione.
            statusTextBlock.Text = "Avvicina il telefono per effettuare il check-in";
            PeerFinder.Start();
        }
    }
}