﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Networking.Proximity;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using Microsoft.WindowsAzure.MobileServices;

// The Basic Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=234237

namespace EasyCheckin.WindowsStore
{
    /// <summary>
    /// A basic page that provides characteristics common to most applications.
    /// </summary>
    public sealed partial class EventPage : EasyCheckin.WindowsStore.Common.LayoutAwarePage
    {
        private bool isConnected;

        private StreamSocket socket;
        private DataReader reader;
        private DataWriter writer;

        private long eventID;

        public EventPage()
        {
            this.InitializeComponent();
        }

        /// <summary>
        /// Populates the page with content passed during navigation.  Any saved state is also
        /// provided when recreating a page from a prior session.
        /// </summary>
        /// <param name="navigationParameter">The parameter value passed to
        /// <see cref="Frame.Navigate(Type, Object)"/> when this page was initially requested.
        /// </param>
        /// <param name="pageState">A dictionary of state preserved by this page during an earlier
        /// session.  This will be null the first time a page is visited.</param>
        protected async override void LoadState(Object navigationParameter, Dictionary<String, Object> pageState)
        {
            // Recupera l'ID dell'evento passato alla pagina come argomento.
            eventID = long.Parse(navigationParameter.ToString());

            // Recupera i dettagli dell'evento.
            var currentEvent = (await (from e in App.MobileService.GetTable<Event>()
                                       where e.Id == eventID
                                       select e).ToListAsync()).FirstOrDefault();

            eventNameTextBox.Text = currentEvent.Name;

            // Recupera le registrazioni attuali.
            this.RefreshRegistrations();

            // Controlla se il device supporta la modalità tap-to-connect.
            if ((PeerFinder.SupportedDiscoveryTypes & PeerDiscoveryTypes.Triggered) != 0)
            {
                PeerFinder.TriggeredConnectionStateChanged += PeerFinder_TriggeredConnectionStateChanged;
                PeerFinder.Start();
            }
            else
            {
                statusTextBlock.Text = "Modalità tap-to-connect non supportata. Impossibile verificare le registrazioni.";
            }
        }

        protected override void OnNavigatedFrom(NavigationEventArgs e)
        {
            // Quando si esce dalla pagina, chiude gli eventuali socket aperti.
            this.CloseSocket();

            PeerFinder.Stop();
            PeerFinder.TriggeredConnectionStateChanged -= PeerFinder_TriggeredConnectionStateChanged;

            base.OnNavigatedFrom(e);
        }

        private void RefreshRegistrations()
        {
            // Recupera la lista degli iscritti ordinati per nome.
            var registrations = (from r in App.MobileService.GetTable<Registration>()
                                 where r.EventId == eventID && r.Confirmed
                                 orderby r.Name
                                 select r).ToIncrementalLoadingCollection();

            // Mostra gli iscritti nella ListView.
            registrationsListView.ItemsSource = registrations;
        }

        private async void PeerFinder_TriggeredConnectionStateChanged(object sender, TriggeredConnectionStateChangedEventArgs args)
        {
            await Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
            {
                switch (args.State)
                {
                    case TriggeredConnectState.PeerFound:   // Il peer è stato trovato, è possibile allontanare i device.
                        statusTextBlock.Text = "Telefono identificato";
                        break;    
                    
                    case TriggeredConnectState.Listening:   // In ascolto.
                        break;

                    case TriggeredConnectState.Connecting:  // Creazione della connessione via Socket.
                        break;

                    case TriggeredConnectState.Completed:   // La connessione è stabilita ed è disponibile.
                        isConnected = true;
                        PeerFinder.Stop();
                        statusTextBlock.Text = "Connessione stabilita";

                        this.HandleConnection(args.Socket);
                        break;

                    case TriggeredConnectState.Canceled:    // La connessione è stata annulla.
                        break;

                    case TriggeredConnectState.Failed:  // Impossibile stabilire la connessione.
                        break;

                    default:
                        break;
                }
            });
        }

        private void HandleConnection(StreamSocket streamSocket)
        {
            // Configura il socket.
            socket = streamSocket;
            reader = new DataReader(streamSocket.InputStream) { InputStreamOptions = InputStreamOptions.None };
            writer = new DataWriter(streamSocket.OutputStream);

            this.ListenAsync(reader);
        }

        private async void ListenAsync(DataReader socketReader)
        {
            try
            {
                // Controlla se lo stream contiene dati, verificando la presenza di un numero che contiene
                // la lunghezza del testo.
                var bytesRead = await socketReader.LoadAsync(sizeof(uint));
                if (bytesRead > 0)
                {
                    // Legge il numero di caratteri presenti nello stream e, subito dopo, carica dallo stream
                    // il numero di byte specificati.
                    var len = socketReader.ReadUInt32();
                    bytesRead = await socketReader.LoadAsync(len);
                    if (bytesRead > 0)
                    {
                        // Se effettivamente ci sono dati, li legge in formato stringa.
                        var message = socketReader.ReadString(len);

                        // Richiama il metodo che gestisce il messaggio ricevuto.
                        var continueRead = await this.HandleMessageAsync(message);

                        if (continueRead)
                        {
                            // Sono previsti ulteriori messaggi, quindi avvia una nuova lettura.
                            this.ListenAsync(socketReader);
                        }
                        else
                        {
                            // La connessione è terminata.
                            this.CloseSocket();
                        }
                    }
                    else
                    {
                        // Il socket remoto è stato chiuso.
                        this.CloseSocket();
                    }
                }
                else
                {
                    this.CloseSocket();
                }
            }
            catch
            {
                this.CloseSocket();
            }
        }

        private async Task<bool> HandleMessageAsync(string message)
        {
            // Controlla se è stato ricevuto l'ID della registrazione.
            long registrationID = 0;
            if (long.TryParse(message, out registrationID))
            {
                var registration = (await (from r in App.MobileService.GetTable<Registration>()
                                           where r.EventId == eventID && r.Id == registrationID
                                           select r).ToListAsync()).FirstOrDefault();
                if (registration != null)
                {
                    // Conferma la registrazione.
                    registration.Confirmed = true;
                    await App.MobileService.GetTable<Registration>().UpdateAsync(registration);

                    statusTextBlock.Text = registration.Name;
                    await Task.Delay(1000);
                    await this.SendAsync("Iscrizione confermata per: " + registration.Name);

                    // Aggiorna le registrazioni.
                    this.RefreshRegistrations();

                    return false;
                }
                else
                {
                    var errorMessage = "Registrazione non valida";
                    statusTextBlock.Text = errorMessage;
                    await Task.Delay(1000);
                    await this.SendAsync(errorMessage);

                    return false;
                }
            }

            return true;
        }

        private async Task SendAsync(string message)
        {
            if (isConnected && !string.IsNullOrEmpty(message))
            {
                try
                {
                    // Calcola il numero di byte da inviare.
                    var len = writer.MeasureString(message);
                    writer.WriteUInt32(len);

                    // Invia il messaggio vero e proprio.
                    writer.WriteString(message);
                    var bytesWritten = await writer.StoreAsync();
                    if (bytesWritten == 0)
                    {
                        // Il socket remoto è stato disconnesso.
                        this.CloseSocket();
                    }
                }
                catch
                {
                    this.CloseSocket();
                }
            }
            else
            {
                this.CloseSocket();
            }
        }

        private void CloseSocket()
        {
            if (reader != null)
            {
                reader.Dispose();
                reader = null;
            }

            if (writer != null)
            {
                writer.Dispose();
                writer = null;
            }

            if (socket != null)
            {
                socket.Dispose();
                socket = null;
            }

            isConnected = false;

            // Chiuso il socket, si mette in attesa di una connessione.
            statusTextBlock.Text = "Avvicina il telefono per effettuare il check-in";
            PeerFinder.Start();
        }
    }
}
